# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC, abstractmethod
from typing import List


class ConversationFormatter(ABC):
    """
    Represents a way of formatting a conversation with an LLM
    such that it can response appropriately
    """

    @abstractmethod
    def format_conversation(self, conv: List[dict]) -> str:
        raise NotImplementedError(
            "format_converstaion must be implemented by subclasses"
        )
